use std::collections::HashMap;

use banner_pattern::BannerPattern;
use chat_type::ChatType;
use damage_type::DamageType;
use dimension_type::DimensionType;
use painting_variant::PaintingVariant;
use thiserror::Error;
use trim_material::TrimMaterial;
use trim_pattern::TrimPattern;
use wolf_variant::WolfVariant;
use worldgen::biome::Biome;

use crate::{TagOrIdentifier, identifier::Identifier, tag::Tag};

pub mod banner_pattern;
pub mod chat_type;
pub mod damage_type;
pub mod dimension_type;
pub mod painting_variant;
pub mod trim_material;
pub mod trim_pattern;
pub mod wolf_variant;
pub mod worldgen;

#[derive(Default)]
pub struct Registries {
    // advancement
    pub banner_patterns: Registry<BannerPattern>,
    pub chat_types: Registry<ChatType>,
    pub damage_types: Registry<DamageType>,
    // dimension
    pub dimension_types: Registry<DimensionType>,
    // enchantment
    // enchantment_provider
    // instrument
    // item_modifier
    // jukebox_song
    // loot_table
    pub painting_variants: Registry<PaintingVariant>,
    // predicate
    // recipe
    // trail_spawner
    pub trim_materials: Registry<TrimMaterial>,
    pub trim_patterns: Registry<TrimPattern>,
    pub wolf_variants: Registry<WolfVariant>,
    pub worldgen: WorldgenRegistries,
}

impl Registries {
    pub fn empty() -> Registries {
        Self::default()
    }
}

#[derive(Default)]
pub struct WorldgenRegistries {
    pub biomes: Registry<Biome>,
    // configured_carver
    // configured_feature
    // density_function
    // noise
    // noise_settings
    // placed_feature
    // processor_list
    // structure
    // structure_set
    // template_pool
    // world_preset
    // flat_level_generator_preset
    // multi_noise_biome_source_parameter_list
}

#[derive(Debug, Error)]
pub enum RegistryError {
    #[error("Duplicate key: {0}")]
    DuplicateKey(Identifier),
}

pub struct Registry<T>
where
    T: Clone,
{
    // Int id -> T
    by_id: Vec<T>,
    // Int id -> Identifier
    id_to_key: Vec<Identifier>,
    // Identifier -> Int Id
    to_id: HashMap<Identifier, usize>,
    // Identifier -> T
    by_key: HashMap<Identifier, T>,

    tags: HashMap<Tag, Vec<TagOrIdentifier>>,
}

impl<T> Default for Registry<T>
where
    T: Clone,
{
    fn default() -> Self {
        Self {
            by_id: Default::default(),
            id_to_key: Default::default(),
            to_id: Default::default(),
            by_key: Default::default(),
            tags: Default::default(),
        }
    }
}

impl<T> Registry<T>
where
    T: Clone,
{
    pub fn register(&mut self, key: Identifier, value: T) -> Result<(), RegistryError> {
        if self.to_id.contains_key(&key) {
            return Err(RegistryError::DuplicateKey(key));
        }
        let id = self.by_id.len();
        self.by_id.push(value.clone());
        self.id_to_key.push(key.clone());
        self.to_id.insert(key.clone(), id);
        self.by_key.insert(key, value);

        Ok(())
    }

    pub fn register_tag(&mut self, tag: Tag, entries: Vec<TagOrIdentifier>) {
        match self.tags.get_mut(&tag) {
            Some(values) => {
                values.extend(entries);
            }
            None => {
                self.tags.insert(tag, entries);
            }
        }
    }

    pub fn replace_tag(&mut self, tag: Tag, entries: Vec<TagOrIdentifier>) {
        self.tags.insert(tag, entries);
    }

    pub fn len(&self) -> usize {
        self.by_id.len()
    }

    pub fn tags_len(&self) -> usize {
        self.tags.len()
    }

    pub fn is_empty(&self) -> bool {
        self.by_id.is_empty()
    }

    pub fn iter(&self) -> impl Iterator<Item = (&Identifier, &T)> {
        self.id_to_key.iter().zip(self.by_id.iter())
    }
}
