use potato_data::{
    Either,
    identifier::Identifier,
    registry::{
        self,
        banner_pattern::BannerPattern,
        chat_type::ChatType,
        damage_type::DamageType,
        dimension_type::DimensionType,
        painting_variant::PaintingVariant,
        trim_material::TrimMaterial,
        trim_pattern::TrimPattern,
        wolf_variant::WolfVariant,
        worldgen::biome::{
            BiomeAdditionsSound, BiomeGrassColorModifier, BiomeMoodSound, BiomeMusicEntry,
            BiomeParticle, BiomeTemperatureModifier, SoundEvent,
        },
    },
};
use potato_protocol_derive::{Packet, PacketEncodable};
use serde::{Deserialize, Serialize};

use crate::packet_encodable::Nbt;

#[derive(Debug, Packet)]
#[packet(configuration_id = crate::ids::configuration::clientbound::REGISTRY_DATA)]
pub struct RegistryDataPacket {
    pub registry_id: Identifier,
    pub entries: Vec<RegistryDataEntry>,
}

#[derive(Debug, PacketEncodable)]
pub struct RegistryDataEntry {
    pub id: Identifier,
    pub data: Option<Nbt<RegistryData>>,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum RegistryData {
    TrimMaterial(Box<TrimMaterial>),
    TrimPattern(Box<TrimPattern>),
    BannerPattern(Box<BannerPattern>),
    Biome(Box<Biome>),
    ChatType(Box<ChatType>),
    DamageType(Box<DamageType>),
    DimensionType(Box<DimensionType>),
    WolfVariant(Box<WolfVariant>),
    PaintingVariant(Box<PaintingVariant>),
}

macro_rules! impl_from_registry {
    ($($variant:ident),*) => {
        $(
            impl From<&$variant> for RegistryData {
                fn from(value: &$variant) -> Self {
                    RegistryData::$variant(Box::new(value.clone()))
                }
            }
        )*
    };
}

impl_from_registry!(
    TrimMaterial,
    TrimPattern,
    BannerPattern,
    ChatType,
    DamageType,
    DimensionType,
    WolfVariant,
    PaintingVariant
);

#[derive(Debug, Serialize, Deserialize)]
pub struct Biome {
    has_precipitation: bool,
    temperature: f32,
    temperature_modifier: BiomeTemperatureModifier,
    downfall: f32,
    effects: BiomeEffects,
}

impl From<&registry::worldgen::biome::Biome> for RegistryData {
    fn from(value: &registry::worldgen::biome::Biome) -> Self {
        RegistryData::Biome(Box::new(Biome {
            has_precipitation: value.has_precipitation,
            temperature: value.temperature,
            temperature_modifier: value.temperature_modifier,
            downfall: value.downfall,
            effects: (&value.effects).into(),
        }))
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct BiomeEffects {
    fog_color: i32,
    water_color: i32,
    water_fog_color: i32,
    sky_color: i32,
    foliage_color: Option<i32>,
    grass_color: Option<i32>,
    grass_color_modifier: BiomeGrassColorModifier,
    particle: Option<BiomeParticle>,
    ambient_sound: Option<Either<Identifier, SoundEvent>>,
    mood_sound: Option<BiomeMoodSound>,
    additions_sound: Option<BiomeAdditionsSound>,
    music: Option<Vec<BiomeMusicEntry>>,
}

impl From<&registry::worldgen::biome::BiomeEffects> for BiomeEffects {
    fn from(value: &registry::worldgen::biome::BiomeEffects) -> Self {
        BiomeEffects {
            fog_color: value.fog_color,
            water_color: value.water_color,
            water_fog_color: value.water_fog_color,
            sky_color: value.sky_color,
            foliage_color: value.foliage_color,
            grass_color: value.grass_color,
            grass_color_modifier: value.grass_color_modifier,
            particle: value.particle.clone(),
            ambient_sound: value.ambient_sound.clone(),
            mood_sound: value.mood_sound.clone(),
            additions_sound: value.additions_sound.clone(),
            music: value.music.clone(),
        }
    }
}
